;nyquist plug-in
;version 1
;type analyze
;name "DTMF Analyzer..."
;action "Finding DTMF tones..."
;info "by Russ Price"

;control mfthresh "Detection threshold versus peak" int "%" 37 5 100
;control delay "Delay threshold" int "ms" 8 1 50
;control dbg "Debug output" int "on/off" 0 0 1

; DTMF analyzer. This is meant to be used with
; recordings like the Evan Doorbell tapes.

; This is a filter-bank based detector which attempts
; to clean up the signal before trying to determine
; the tone values.

; Copyright 2018 by Russ Price <cnet@octothorpe.info>

; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.

; Standalone debugging support.
(setq standalone NIL)			; Are we an Audacity plugin or standalone?
(if standalone (setq dbg T) (if (= dbg 1) (setq dbg T) (setq dbg NIL)))

(if standalone (setq *gc-flag* NIL))	; Make standalone Nyquist shut up.

(if standalone (setq delay 40))		; milliseconds
(if standalone (setq mfthresh 45))	; detection threshold for recombined DTMF

; Read in raw DTMF if standalone.
(if standalone (setf s (s-read "./rawdtmf.wav")))

; DTMF frequency table.
(setq freqtbl (list
	(list "1" 697 1209)
	(list "2" 697 1336)
	(list "3" 697 1477)
	(list "A" 697 1633)
	(list "4" 770 1209)
	(list "5" 770 1336)
	(list "6" 770 1477)
	(list "B" 770 1633)
	(list "7" 852 1209)
	(list "8" 852 1336)
	(list "9" 852 1477)
	(list "C" 852 1633)
	(list "*" 941 1209)
	(list "0" 941 1336)
	(list "#" 941 1477)
	(list "D" 941 1633)
))

; Apply bandpass over entire frequency range +- 50 Hz.
(setf sbp (reson s 1165 986 2))

; Filter sound; don't normalize for DTMF.
; Two passes of biquad with q=20 works wonders.
(defun dofilter (fsnd freq)
	(setf pass1 (bandpass2 fsnd freq 20))
	(setf pass2 (bandpass2 pass1 freq 20))
	pass2
)

; Filter and normalize each DTMF frequency.
(setf s697 (dofilter sbp 697))
(setf s770 (dofilter sbp 770))
(setf s852 (dofilter sbp 852))
(setf s941 (dofilter sbp 941))
(setf s1209 (dofilter sbp 1209))
(setf s1336 (dofilter sbp 1336))
(setf s1477 (dofilter sbp 1477))
(setf s1633 (dofilter sbp 1633))

; Recombine filtered frequencies into a cleaned-up DTMF sequence.
(setf filtered (snd-add s697 (snd-add s770 (snd-add s852 (snd-add s941 (snd-add s1209 (snd-add s1336 (snd-add s1477 s1633))))))))
;(play filtered)

; Get a signal envelope from a sound.
(defun signal (signalsnd) (force-srate 1000 (lp (snd-follow (lp signalsnd 50) 0.001 0.01 0.1 512) 10)))

; Find beats and return a list of time values.
; These are returned in reverse order.
; fbsig: signal envelope
; thresval: threshold to count a beat
; maxi: peak signal value
; WARNING: This destroys the signal envelope!
(defun fb (fbsig thresval maxi)
	(setq thres (* (/ thresval 100.0) maxi))
	(do ((c 0.0) (l NIL) (tval 0) (v (snd-fetch fbsig))) ((not v) l)
		(if dbg (list
			(setq pct (* (/ v maxi) 100.0))
;			(if (> v thres) (display c v thres pct 'HI) (display c v thres pct 'LO)))
			(if (> v thres) (format *standard-output* "~A,~A,~A,~A~%" c pct tval "HI") (format *standard-output* "~A,~A,~A,~A~%" c pct tval "LO")))
		)
		(if (> v thres) (setq tval (1+ tval)) (setq tval 0))
		(if (= tval delay) (list (if dbg (print "BEEP!")) (setq l (cons c l))))
		(setq c (+ c 0.001))
		(setq v (snd-fetch fbsig))
	)
)

; Determine if level at a certain time exceeds the threshold.
(defun isactive (samples sampno thresval maxi len)
	(setq val (aref samples sampno))
	(if dbg (list
		(setq pct (* (/ val maxi) 100.0))
		(display sampno val pct thresval))
	)
	(>= val (* (/ thresval 100.0) maxi))
)
	
; Get the DTMF key value for a pair of frequencies.
(defun getkey (freqs)
	(setq rval NIL)
	(dolist (elem freqtbl)
		(if (equal freqs (cdr elem)) (setq rval (car elem)))
	)
	rval
)

; Needed for sorting list of 2-item lists.
(defun cdrgt (q r) (> (car (cdr q)) (car (cdr r))))

; Snag signal envelopes
(setf e697 (signal s697))
(setf e770 (signal s770))
(setf e852 (signal s852))
(setf e941 (signal s941))
(setf e1209 (signal s1209))
(setf e1336 (signal s1336))
(setf e1477 (signal s1477))
(setf e1633 (signal s1633))
(setf efiltered (signal filtered))

; Get our envelope length (it's the same for all envelopes)
(setf slen (snd-length efiltered NY:ALL))

; Get envelope peaks
(setf p697 (peak e697 NY:ALL))
(setf p770 (peak e770 NY:ALL))
(setf p852 (peak e852 NY:ALL))
(setf p941 (peak e941 NY:ALL))
(setf p1209 (peak e1209 NY:ALL))
(setf p1336 (peak e1336 NY:ALL))
(setf p1477 (peak e1477 NY:ALL))
(setf p1633 (peak e1633 NY:ALL))
(setf pfiltered (peak efiltered NY:ALL))

; Get envelope arrays
(setf a697 (snd-samples e697 slen))
(setf a770 (snd-samples e770 slen))
(setf a852 (snd-samples e852 slen))
(setf a941 (snd-samples e941 slen))
(setf a1209 (snd-samples e1209 slen))
(setf a1336 (snd-samples e1336 slen))
(setf a1477 (snd-samples e1477 slen))
(setf a1633 (snd-samples e1633 slen))

(setf beats (reverse (fb efiltered mfthresh pfiltered)))

(if dbg (print beats))

(setq mflist NIL)
(dolist (mark beats)
	(setq sample (truncate (* mark 1000)))
	(setf flist (list
		(list 697 (aref a697 sample))
		(list 770 (aref a770 sample))
		(list 852 (aref a852 sample))
		(list 941 (aref a941 sample))
		(list 1209 (aref a1209 sample))
		(list 1336 (aref a1336 sample))
		(list 1477 (aref a1477 sample))
		(list 1633 (aref a1633 sample)))
	)
	; Sort frequency list by descending magnitude.
	(setf slist (sort flist 'cdrgt))
	; Take top 2 frequencies.
	(setq rawtop2 (list (car (first slist)) (car (second slist))))
	; Sort them in ascending order.
	(setq top2 (sort rawtop2 '<))
	(setq key (getkey top2))
	(if dbg (display mark slist top2 key))
	(if key (setq mflist (cons (list mark key) mflist)))
)
(if standalone (display 'final mflist))

(if standalone (exit) mflist)
