;nyquist plug-in
;version 1
;type analyze
;name "MF Analyzer..."
;action "Finding MF tones..."
;info "by Russ Price"

;control mfthresh "Detection threshold versus peak" int "%" 55 5 100
;control delay "Delay threshold" int "ms" 15 1 50
;control norm "Normalize each frequency" int "on/off" 1 0 1
;control dbg "Debug output" int "on/off" 0 0 1

; MF analyzer. This is meant to be used with
; recordings like the Evan Doorbell tapes.

; This is a filter-bank based detector which attempts
; to clean up the signal before trying to determine
; the tone values.

; Copyright 2018 by Russ Price <cnet@octothorpe.info>

; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.

; Standalone debugging support.
(setq standalone NIL)			; Are we an Audacity plugin or standalone?
(if standalone (setq dbg T) (if (= dbg 1) (setq dbg T) (setq dbg NIL)))

(if standalone (setq *gc-flag* NIL))	; Make standalone Nyquist shut up.

(if standalone (setq delay 40))		; milliseconds
(if standalone (setq mfthresh 45))	; detection threshold for recombined MF
(if standalone (setq norm 0))

; Read in raw MF.
(if standalone (setf s (s-read "mf.wav")))

; MF frequency table.
(setq freqtbl (list
	(list "1"    700  900)
	(list "2"    700 1100)
	(list "3"    900 1100)
	(list "4"    700 1300)
	(list "5"    900 1300)
	(list "6"   1100 1300)
	(list "7"    700 1500)
	(list "8"    900 1500)
	(list "9"   1100 1500)
	(list "0"   1300 1500)
	(list "ST3"  700 1700)
	(list "ST2"  900 1700)
	(list "KP"  1100 1700)
	(list "KP2" 1300 1700)
	(list "ST"  1500 1700)
))

; Cleanup cruft.
(setf sbp (reson s 1200 1100 2))
(areson sbp 800 100 2)
(areson sbp 1000 100 2)
(areson sbp 1200 100 2)
(areson sbp 1400 100 2)
(areson sbp 1600 100 2)

; Normalize sound according to peak.
(defun scalesound (ssnd)
	(setq pk (peak ssnd NY:ALL))
	(setq scalefactor (/ 1 pk))
	(snd-scale scalefactor ssnd)
)

; Filter and normalize a sound.
; Two passes of biquad with q=20 works wonders.
(defun dofilter (fsnd freq)
	(setf pass1 (bandpass2 fsnd freq 20))
	(setf pass2 (bandpass2 pass1 freq 20))
	(if (= norm 1) (scalesound pass2) pass2)
)

; Filter and normalize each MF frequency.
; By normalizing each frequency, we can cope
; with the badly twisted MFs on crossbar tandems.
(setf s700 (dofilter sbp 700))
(setf s900 (dofilter sbp 900))
(setf s1100 (dofilter sbp 1100))
(setf s1300 (dofilter sbp 1300))
(setf s1500 (dofilter sbp 1500))
(setf s1700 (dofilter sbp 1700))

; Recombine filtered frequencies into a cleaned-up MF sequence.
(setf filtered (snd-add s700 (snd-add s900 (snd-add s1100 (snd-add s1300 (snd-add s1500 s1700))))))
;(play filtered)

; Get a signal envelope from a sound.
(defun signal (signalsnd) (force-srate 1000 (lp (snd-follow signalsnd 0.001 0.01 0.1 512) 10)))

; Find beats and return a list of time values.
; These are returned in reverse order.
; fbsig: signal envelope
; thresval: threshold to count a beat
; maxi: peak signal value
; WARNING: This destroys the signal envelope!
(defun fb (fbsig thresval maxi)
	(setq thres (* (/ thresval 100.0) maxi))
	(do ((c 0.0) (l NIL) (tval 0) (v (snd-fetch fbsig))) ((not v) l)
		(if dbg (list
			(setq pct (* (/ v maxi) 100.0))
;			(if (> v thres) (display c v thres pct 'HI) (display c v thres pct 'LO)))
			(if (> v thres) (format *standard-output* "~A,~A,~A,~A~%" c pct tval "HI") (format *standard-output* "~A,~A,~A,~A~%" c pct tval "LO")))
		)
		(if (> v thres) (setq tval (1+ tval)) (setq tval 0))
		(if (= tval delay) (list (if dbg (print "BEEP!")) (setq l (cons c l))))
		(setq c (+ c 0.001))
		(setq v (snd-fetch fbsig))
	)
)

; Determine if level at a certain time exceeds the threshold.
(defun isactive (samples sampno thresval maxi len)
	(setq val (aref samples sampno))
	(if dbg (list
		(setq pct (* (/ val maxi) 100.0))
		(display sampno val pct thresval))
	)
	(>= val (* (/ thresval 100.0) maxi))
)
	
; Get the MF key value for a pair of frequencies.
(defun getkey (freqs)
	(setq rval NIL)
	(dolist (elem freqtbl)
		(if (equal freqs (cdr elem)) (setq rval (car elem)))
	)
	rval
)

; Needed for sorting list of 2-item lists.
(defun cdrgt (q r) (> (car (cdr q)) (car (cdr r))))

; Snag signal envelopes
(setf e700 (signal s700))
(setf e900 (signal s900))
(setf e1100 (signal s1100))
(setf e1300 (signal s1300))
(setf e1500 (signal s1500))
(setf e1700 (signal s1700))
(setf efiltered (signal filtered))

; Get our envelope length (it's the same for all envelopes)
(setf slen (snd-length efiltered NY:ALL))

; Get envelope peaks
(setf pfiltered (peak efiltered NY:ALL))

; Get envelope arrays
(setf a700 (snd-samples e700 slen))
(setf a900 (snd-samples e900 slen))
(setf a1100 (snd-samples e1100 slen))
(setf a1300 (snd-samples e1300 slen))
(setf a1500 (snd-samples e1500 slen))
(setf a1700 (snd-samples e1700 slen))

(setf beats (reverse (fb efiltered mfthresh pfiltered)))

(if dbg (print beats))

(setq mflist NIL)
(dolist (mark beats)
	(setq sample (truncate (* mark 1000)))
	(setf flist (list
		(list 700 (aref a700 sample))
		(list 900 (aref a900 sample))
		(list 1100 (aref a1100 sample))
		(list 1300 (aref a1300 sample))
		(list 1500 (aref a1500 sample))
		(list 1700 (aref a1700 sample)))
	)
	; Sort frequency list by descending magnitude.
	(setf slist (sort flist 'cdrgt))
	; Take top 2 frequencies.
	(setq rawtop2 (list (car (first slist)) (car (second slist))))
	; Sort them in ascending order.
	(setq top2 (sort rawtop2 '<))
	(setq key (getkey top2))
	(if dbg (display mark slist top2 key))
	(if key (setq mflist (cons (list mark key) mflist)))
)
(if standalone (display 'final mflist))
(if standalone () mflist)
